#! /bin/bash
set -eu
ROOT_DIR=$(dirname $(readlink -f $0))

log_debug() {
    [[ -L /dev/log || -f /dev/log ]] && logger -p local3.debug -t build_system[$$] $@
    echo "[DEBUG] $@"
}
log_info() {
    [[ -L /dev/log || -f /dev/log ]] && logger -p local3.info -t build_system[$$] $@
    echo "[INFO] $@"
}
log_warning() {
    [[ -L /dev/log || -f /dev/log ]] && logger -p local3.warning -t build_system[$$] $@
    echo "[WARN] $@"
}
log_error() {
    [[ -L /dev/log || -f /dev/log ]] && logger -p local3.err -t build_system[$$] $@
    echo "[ERROR] $@" >&2
}

# 判断系统版本是否符合
system_version=$(lsb_release -rs)
if [ "$(echo "$system_version != 16.04" | bc)" -ne 0 ]; then
    log_info "The system version does not match the environment package version. Please run the script in the xubuntu 16.04 environment."
    exit 1
fi

if grep 'local3.* /var/log/arcs_system.log' /etc/syslog.conf &>/dev/null;then
    log_info "The syslog config already exists"
else
    log_info "Add syslog config"
    echo -e "\nlocal3.* /var/log/arcs_system.log" >> /etc/syslog.conf
    systemctl restart syslog ||:
fi


[ -f /etc/default/arcs ] && source /etc/default/arcs

[ -f ${ROOT_DIR}/version ] && source ${ROOT_DIR}/version

VERSION=${VERSION:-0.0.0}

log_info start installing version ${VERSION} in arcs environment...

if ! [ $(id -u) = 0 ]; then
    log_info "$0: The script need to be run as root."
    exit 1
fi

# 0安装
# 0.1触摸屏驱动(注意要等待执行结束,最后生成触摸屏配置数据库文件)
install_pmLinux(){
    log_info "install_pmLinux.."
    cd ${ROOT_DIR}
    tar -jxvf pmLinux-Ubuntu16.04_v4_7_6_R43_x86_64_20230629.tar.bz2
    cd ./pmLinux-Ubuntu16.04
    log_info $(./install.sh)
    cd ${ROOT_DIR}
    # rm -rf pmLinux-Ubuntu
}

# 0.2 安装rtl8812au无线驱动
install_rtl8812au(){
    echo "install_rtl8812au..."
    insmod /lib/modules/4.14.195-rt94/updates/dkms/* || true
}

# 0.5 拷贝libstdc++.so.6,libxkbcommon库文件
copy_files() {
    log_info "copy_files.."
    chown -R root:root chroot/
    rsync -a -r chroot/  /
    chmod +x /usr/bin/*
}

# 保持主机名一致
fix_hosts() {
    log_info "fix_hosts.."
    new_hostname=$(cat /etc/hostname)
    new_ip="127.0.1.1"
    sed -i "2s/.*/$new_ip\t$new_hostname/" /etc/hosts
}

clean_srs_and_aubope() {
    rm -f /etc/rc.local
    test -d /root/auborobotstudio && rm -rf /root/auborobotstudio
    test -f /root/aubostart && rm -f /usr/bin/aubostart
    sed -i '/aubostart/d' /etc/profile
}

## NOTE: 必须联网安装 xubuntu-desktop
install_xubuntu_desktop() {
    log_info "Configure offline installation source"
    # 配置离线安装源
    echo "deb file:///${ROOT_DIR}/offline-packages archives/"| tee /etc/apt/sources.list
    
    mv /var/lib/dpkg/info/ /var/lib/dpkg/info_old/
    mkdir /var/lib/dpkg/info/
    apt-get update
    apt install xubuntu-desktop -y
    mv /var/lib/dpkg/info/* /var/lib/dpkg/info_old/
    rm -rf /var/lib/dpkg/info
    mv /var/lib/dpkg/info_old/ /var/lib/dpkg/info/
}

cpu_iso() {
    log_info "cpu_iso.."
    # 1实时性配置
    # 1.1 孤立cpu1
    sed -i 's/quiet splash/quiet splash isolcpus=1/' /etc/default/grub
    update-grub
}

cpu_affinity() {
    log_info "cpu_affinity.."
    systemctl stop irqbalance.service

    no_rt_kernerl=3
    # 1.2 更改中断的CPU亲和(终端不显示错误信息,因为有些终端不允许移动到别的核)
    for file in `find /proc/irq -name "smp_affinity_list"`; do
        echo ${no_rt_kernerl} > ${file} > /dev/null;
    done
    
    # `KWorker`的CPU亲和(Linux kernel中将工作推后执行的一种机制)
    for file in `find /sys/devices/virtual/workqueue -name "cpumask"`; do
        echo ${no_rt_kernerl} > ${file} 2>/dev/null;
    done
    echo ${no_rt_kernerl} > /sys/devices/virtual/workqueue/cpumask
}

cpu_freq() {
    log_info "cpu_afcpu_freqfinity.."
    # 1.3 禁用cpu调频
    
    # 将cpu设置为性能模式
    np=`cat /proc/cpuinfo | grep processor | wc -l`
    let num=($np -1)
    for i in $(seq 0 $num)
    do
        cpufreq-set -c $i -g performance
    done
    cpufreq-info
}

disable_culling() {
    log_info "disable_culling.."
    # 2.解决鼠标图标重影问题
    echo "CLUTTER_PAINT=disable-clipped-redraws:disable-culling" >> /etc/environment
}

sshd_config() {
    log_info "sshd_config.."
    # 3.配置默认ssh端口 为8822, 允许 root 远程登录
    sed -i 's/Port 22/Port 8822/;s/PermitRootLogin prohibit-password/PermitRootLogin yes/' /etc/ssh/sshd_config
}

root_auto_login() {
    log_info "root_auto_login.."
    # 4.修改用户密码
    echo 'root:bestcobot' | chpasswd
    
    # 5.配置 root 自动登录
    test -f /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf && rm /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "[Seat:*]" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "user-session=xubuntu" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "autologin-guest=false" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "autologin-user=root" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "autologin-user-timeout=0" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    echo "greeter-session=lightdm-gtk-greeter" >> /usr/share/lightdm/lightdm.conf.d/60-xubuntu.conf
    
    test -f /etc/lightdm/lightdm.conf && rm /etc/lightdm/lightdm.conf
    echo "[Seat:*]" >> /etc/lightdm/lightdm.conf
    echo "autologin-session=xubuntu" >> /etc/lightdm/lightdm.conf
    echo "autologin-guest=false" >> /etc/lightdm/lightdm.conf
    echo "autologin-user=root" >> /etc/lightdm/lightdm.conf
    echo "autologin-user-timeout=0" >> /etc/lightdm/lightdm.conf
    
    # 修改/root/.profile文件
    sed -i 's/mesg n/tty -s \&\& mesg n/' /root/.profile
}

xinput_map() {
    log_info "xinput_map.."
    # 6.开机自启动映射 xinput (显示器和触摸屏先建立映射关系)
    mkdir -p /root/.config/autostart
cat <<EOF > /root/.config/autostart/xinput.desktop
[Desktop Entry]
Type=Application
Exec=xfce4-terminal -e bash -c 'xinput map-to-output "DIALOGUE INC PenMount USB" eDP-1'
Hidden=false
NoDisplay=false
X-GNOME-Autostart-enabled=true
EOF
}

disable_update_and_apport() {
    log_info "disable_update_and_apport.."
    # 8.关闭系统崩溃提示,关闭系统自动更新升级，删除更新通知程序
    sed -i "s/enabled=1/enabled=0/g" /etc/default/apport
    sed -i "s/1/0/g" /etc/apt/apt.conf.d/10periodic
    sed -i "s/1/0/g" /etc/apt/apt.conf.d/20auto-upgrades
    sudo apt-get remove update-notifier -y
}

optimize_networking_service(){
    log_info "optimize_networking_service.."
    sed -i "s/TimeoutStartSec=5min/TimeoutStartSec=5s/g" /lib/systemd/system/networking.service
    
    # 禁用系统启动时候等待网络连接的服务
    systemctl disable NetworkManager-wait-online.service
}

config_vnc() {   
    log_info "config_vnc.."
    sudo dpkg -i x11vnc/*.deb

    mypasswd="bestcobot"
    x11vnc -storepasswd $mypasswd /root/.vnc/passwd
    chown -R root /root/.vnc
    chmod 0600 /root/.vnc/passwd

    sudo systemctl daemon-reload
    sudo systemctl enable x11vnc.service
    sudo systemctl start x11vnc.service
}

set_tcp_fin_timeout(){
    log_info "set_tcp_fin_timeout.."
    # 设置端口释放后的等待时间为30s
    sysctl -w net.ipv4.tcp_fin_timeout=30
    # 释放TIME_WAIT端口给新连接使用
    sysctl -w net.ipv4.tcp_timestamps=1
}

# u盘自动挂载识别并执行程序
config_autousb() {
    log_info "config_autousb..."
    chmod a+x /usr/local/bin/usbplug.sh
    systemctl daemon-reload
    systemctl restart udev
}

# 添加系统实时监控
config_monitor() {
    log_info "config_monitor..."
    chmod a+x /usr/local/bin/record_system_info.sh
    systemctl daemon-reload
    systemctl enable system_info_monitor
}

# 添加系统跑分
config_sysbench() {
    log_info "config_sysbench..."
    chmod a+x /usr/local/bin/system-benchmark.sh
}

# 添加exfat
config_exfat() {   
    log_info "config_exfat.."
    sudo dpkg -i exfat/*.deb
}

# 添加7z
config_7zip() {
    log_info "config_7zip.."
    sudo dpkg -i 7z/*.deb
}

# 限制系统日志大小
config_logrotate() {
    log_info "config_logrotate"
    chmod 644 /etc/logrotate.d/rsyslog
    chmod 644 /etc/logrotate.d/monitor
    [[ $(grep logrotate /etc/crontab|wc -l) == "0" ]] && echo "*/2 * * * *   root    /bin/bash /etc/cron.daily/logrotate > /dev/null 2>&1" >> /etc/crontab
}

config_create_ap() {
    cat > /etc/create_ap.conf <<-EOF
CHANNEL=8
GATEWAY=10.0.0.1
WPA_VERSION=2
ETC_HOSTS=0
DHCP_DNS=gateway
NO_DNS=0
NO_DNSMASQ=0
HIDDEN=0
MAC_FILTER=0
MAC_FILTER_ACCEPT=/etc/hostapd/hostapd.accept
ISOLATE_CLIENTS=0
SHARE_METHOD=nat
IEEE80211N=0
IEEE80211AC=0
HT_CAPAB=[HT40+]
VHT_CAPAB=
DRIVER=nl80211
NO_VIRT=0
COUNTRY=CN
FREQ_BAND=2.4
NEW_MACADDR=
DAEMONIZE=0
NO_HAVEGED=0
WIFI_IFACE=wlp3s0
INTERNET_IFACE=enp1s0
SSID=inactivated-$1
PASSPHRASE=12345678
USE_PSK=0
EOF

    WIFIIFACE=$(nmcli device | grep "wifi[ ]" | awk '{print $1}')
    sed -i "s/wlp5s0/${WIFIIFACE}/g" /etc/create_ap.conf
    systemctl enable create_ap
}


clean_srs_and_aubope
dpkg -l xfdesktop4 > /dev/null || install_xubuntu_desktop

copy_files
fix_hosts
cpu_iso
cpu_affinity
# 虚拟机和docker环境下不执行cpu调频脚本
if lscpu | grep "VMware" > /dev/null;then
    log_info "virtual host"
elif [ "${CONTAINER:-unbound}" != "unbound" ];then
    log_info "container"
else 
    cpu_freq
fi

sshd_config
root_auto_login
disable_update_and_apport
optimize_networking_service
config_vnc
set_tcp_fin_timeout
config_autousb
config_monitor
config_sysbench
config_exfat
config_7zip
config_create_ap $(date +%s%N | md5sum | cut -c 1-9)
config_logrotate

# 虚拟机下不执行安装触屏驱动脚本
lscpu | grep "VMware" > /dev/null || (install_pmLinux && xinput_map && disable_culling)
install_rtl8812au

# 移除xfce4-panel
apt remove xfce4-panel -y

log_info "The ARCS software configuration environment is complete !"
log_info "Successful !"

log_info Installation completed version ${VERSION} arcs environment...
